/***********************************************************************
 *
 * FileName:
 *		PnpView.h
 *
 * Description:
 *    Implementation of view functions for "Palm Network Programming"
 *    sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 *    All network error checking code has been retained for purposes of
 *    example.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpView.h"     // PnpViewConnect(), PnpViewDisconnect().
#include "NetMgr.h"     // NetLib API.

UInt32 PrvPnpViewGetAvailMemory (void);

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
const UInt16 kDrawRectLeft = 3;
const UInt16 kDrawRectTop = 18;
const UInt16 kDrawRectWidth = 154;
const UInt16 kDrawRectHeight = 122;
const UInt16 kDrawHorzSpace = 2;
const UInt16 kDrawVertSpace = 2;

static UInt16 sVertPos = 0;         // Vertical position in window.

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpViewClear:
   Clear the viewport.
   Parms:   caption        Caption for viewport.
   Returns: (none)
*/
void
PnpViewClear (Char* caption)
{
   // Declare local variables.
   RectangleType drawRect;
   FontID prevFontID = stdFont;  // Prev. font id.
   UnderlineModeType prevULine = noUnderline;   // Prev. underline mode.
   UInt16 textLength = 0;        // Length of caption (in chars).
   UInt16 textWidth = 0;         // Display width of caption (in pixels).
   UInt16 horzPos = 0;           // Horz. start of caption (in pixels).

   // Populate draw rectangle.
   RctSetRectangle (&drawRect,
                    kDrawRectLeft,
                    kDrawRectTop,
                    kDrawRectWidth,
                    kDrawRectHeight);

   // Clear the draw rectangle.
   WinEraseRectangle (&drawRect, 0);

   // Draw the rectangle frame.
   WinDrawRectangleFrame (simpleFrame, &drawRect);

   // Set current vertical position.
   sVertPos = kDrawRectTop + kDrawVertSpace;

   // Set the font to bold.
   prevFontID = FntSetFont (boldFont);
   
   // Turn on underlining.
   prevULine = WinSetUnderlineMode (solidUnderline);

   // Get length of string.
   textLength = StrLen (caption);

   // Get the width of the string.
   textWidth = FntCharsWidth (caption, textLength);

   // Calc. the x coord for caption.
   horzPos = kDrawRectLeft + (kDrawRectWidth - textWidth) / 2;

   // Draw the caption.
   WinDrawChars (caption, textLength, horzPos, sVertPos);

   // Update the vertical position.
   sVertPos += FntCharHeight () + kDrawVertSpace;

   // Restore previous underline mode.
   WinSetUnderlineMode (prevULine);

   // Restore previous font.
   FntSetFont (prevFontID);

   // Return.
   return;
}

/*
   PnpViewDrawText:
   Draw a line of text on the viewport.
   Parms:   text           Text to draw.
*/
void
PnpViewDrawText (Char* text)
{
   // Declare local variables.
   FontID prevFontID = stdFont;   // Prev. font id.

   // Set the font to normal.
   prevFontID = FntSetFont (stdFont);
   
   // Draw text.
   WinDrawChars (text,
                 StrLen (text),
                 kDrawRectLeft + kDrawHorzSpace,
                 sVertPos);

   // Update the vertical position.
   sVertPos += FntCharHeight ();

   // Restore previous font.
   FntSetFont (prevFontID);

   // Return.
   return;
}

/*
   PnpViewDrawResources:
   Draw the current resources on the viewport.
   Parms:   text           Text to draw with resources.
*/
void
PnpViewDrawResources (Char* text)
{
   // Declare local variables.
   Char* resourceTextP = NULL;   // Resource text string.
   UInt32 availMemory = 0;       // Available memory.

   // Allocate resource text.
   resourceTextP = (Char*) MemPtrNew (100);

   // Check allocation.
   if (!resourceTextP)
   {
   // Allocation failed.

      // Draw error message.
      PnpViewDrawText ("Unable to allocate text buffer!");

      // Return.
      return;
   }

   // Get resources.
   availMemory = PrvPnpViewGetAvailMemory ();

   // Convert bytes to kbytes.
   availMemory = (availMemory + 500L) / 1000L;

   // Check if add'l text requested.
   if (text)
   {
      // Format memory for output.
      StrPrintF (resourceTextP,
                 "%liKB avail [%s]",
                 availMemory,
                 text);
   }
   else
   {
      // Format memory for output.
      StrPrintF (resourceTextP,
                 "%liKB avail",
                 availMemory);
   }

   // Draw resource text.
   PnpViewDrawText (resourceTextP);

   // Release the memory.
   MemPtrFree (resourceTextP);

   // Return.
   return;
}

/*
   PnpViewDrawVSpace:
   Draw the current resources on the viewport.
   Parms:   (none)
*/
void
PnpViewDrawVSpace ()
{
   // Add one vert space to vert position.
   sVertPos += kDrawVertSpace;
}

UInt32
PrvPnpViewGetAvailMemory (void)
{
   // Declare local variables.
   UInt32 availMemory = 0;    // Available memory.
   UInt16 numHeaps = 0;       // Number of heaps to traverse.
   UInt16 i = 0;              // Iterator.
   UInt16 heapID;             // Identifier for each heap.
   UInt32 maxChunk = 0;       // Max chunk size (per heap).
   UInt32 freeMemory = 0;     // Free memory (per heap).

   // Get the number of heaps.
   numHeaps = MemNumHeaps (0);

   // Traverse the heaps.
   for (i = 0; i < numHeaps; i++)
   {
   // For each heap...
 
      // Get the heap identifier.
      heapID = MemHeapID (0, i);

      // Check if the heap is dynamic.
      if (MemHeapDynamic (heapID))
      {
      // Heap is dynamic.

         // Get amount of free memory (and size of max chunk).
         MemHeapFreeBytes (heapID, &freeMemory, &maxChunk);

         // Add free memory
         availMemory += freeMemory;
      }
   }

   // Return value.
   return availMemory;
}

