/***********************************************************************
 *
 * FileName:
 *		PnpFtp.c
 *
 * Description:
 *    Implementation of FTP functions for "Palm Network Programming"
 *    sample FTP Client application.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Include FTP Function declarations.
#include "PnpFtp.h"
#include "PnpNet.h"
#include "PnpLog.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const UInt16 kOutBufSize = 64;
static const UInt16 kInBufSize = 512;
static const UInt16 kDataBufSize = 1024;
static Char  kCRLF [] = { 0x0D, 0x0A, 0x00 };

static Char * pnpFtpTranNames [] =
{
  "USER",
  "PASS",
  "QUIT",
  "PORT",
  "LIST"
};

static UInt16 const pnpFtpTranCount = sizeof (pnpFtpTranNames) / sizeof (pnpFtpTranNames [0]);

static Char*  kDfltUser = "anonymous";
static const Char* kDfltPass = "pnp@gregsprogrammingworks.com";
   
static const UInt16 kFtpMaxQ  = 4;

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static Err     PrvFtpNetConnect (PnpFtpInfoPtr infoP);
static Err     PrvFtpNetDisconnect (PnpFtpInfoPtr infoP);
static Err     PrvFtpConnect (PnpFtpInfoPtr infoP);
static Err     PrvFtpSelect (PnpFtpInfoPtr infoP);
static Err     PrvFtpDisconnect (PnpFtpInfoPtr infoP);
static Err     PrvFtpUpdate (PnpFtpInfoPtr infoP);
static Boolean PrvFtpIsTranSuccess (PnpFtpInfoPtr infoP);
static Err     PrvFtpQueueCommand (PnpFtpInfoPtr infoP, PnpFtpTranCode tranCode, const Char* argP);
static Err     PrvFtpWriteCommand (PnpFtpInfoPtr infoP);
static Err     PrvFtpReadResponses (PnpFtpInfoPtr infoP);
static Int16   PrvFtpReadResponse (PnpFtpInfoPtr infoP);
static Err     PrvFtpReadControl (PnpFtpInfoPtr infoP);
static Err     PrvFtpCloseControl (PnpFtpInfoPtr infoP);
static Err     PrvFtpDataExchange (PnpFtpInfoPtr infoP);
static Err     PrvFtpDataListen (PnpFtpInfoPtr infoP);
static Err     PrvFtpDataAccept (PnpFtpInfoPtr infoP);
static Err     PrvFtpDataReceive (PnpFtpInfoPtr infoP);
static Err     PrvFtpDataComplete (PnpFtpInfoPtr infoP);
static Int16   PrvFtpGetFirstPositiveReply (PnpFtpInfoPtr infoP);
static Err     PrvFtpCalcPortAddress (PnpFtpInfoPtr infoP, Char* portString);
static Err     PrvFtpAddLogItem (PnpFtpInfoPtr infoP, Char* logItem);
static Int16   PrvFtpFindCRLF (const UInt8* bufP, UInt16 size);

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpFtpInit:
   Initialize FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpFtpInit (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Initialize Network library.
   err = PnpNetInit ();
   if (err)
   {
      return -1;
   }

   // Initialize FTP session information.
   infoP->isNetConnect = false;
   
   // Initialize control related attributes.
   infoP->ctrlSocket = kNilSocket;
   infoP->ctrlInBufH = GBufNew (kInBufSize);
   infoP->ctrlOutBufH = GBufNew (kOutBufSize);

   // Check allocations.
   ErrFatalDisplayIf (NULL == infoP->ctrlInBufH, "Error creating inbound buffer.");
   ErrFatalDisplayIf (NULL == infoP->ctrlOutBufH, "Error creating outbound buffer.");

   // Initialize FTP Transaction information.
   err = GQueInit (&infoP->tranQ, kFtpMaxQ);
   infoP->tranCode = kTranCodeInactive;
   infoP->tranState = kTranStateInactive;

   // We will clear the FD items before we call Select.
   
   // Initialize FTP Data socket information.
   infoP->dataSocket = kNilSocket;
   infoP->dataState = kDataStateNoTransfer;
   
   // Log successful initialization.
   PrvFtpAddLogItem (infoP, "INITIALIZE");

   // Return result.
   return 0;
}

/*
   PnpFtpTerm:
   Terminate FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpFtpTerm (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Disconnect from the network.
   // (This will end the FTP session, if necessary)
   PrvFtpNetDisconnect (infoP);
   
   // Check if inbound buffer allocated.
   if (NULL != infoP->ctrlInBufH)
   {
   // Buffer is allocated.

      // Free buffer.
      GBufFree (infoP->ctrlInBufH);
      infoP->ctrlInBufH = NULL;
   }

   // Check if outbound buffer allocated.
   if (NULL != infoP->ctrlOutBufH)
   {
   // Buffer is allocated.

      // Free buffer.
      GBufFree (infoP->ctrlOutBufH);
      infoP->ctrlOutBufH = NULL;
   }

   // Close down the queue.
   GQueTerm (&infoP->tranQ);

   // Terminate network.
   PnpNetTerm ();

   // Log successful initialization.
   PrvFtpAddLogItem (infoP, "TERMINATE");

   // Return result.
   return err;
}

/*
   PnpFtpProcess:
   Process any incoming or outgoing FTP events.
   Params:  infoP    FTP Session information.
   Returns: 0        FTP events handled successfully.
            <0       FTP connection ended.
*/
Err
PnpFtpProcess (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;         // Return value.

   // Check connections.
   err = PrvFtpConnect (infoP);
   
   // Check if we have any connections.
   if (kNilSocket != infoP->ctrlSocket)
   {
   // We are connected!
   
      // Check data exchange.
      err = PrvFtpSelect (infoP);
      
      // Check if there is data to read or write.
      
      if (!err)
      {
      // Sockets are ready. 

         // If we read a response or write a command,
         // we want to return success. Since success
         // is 0, we'll set an error and then and it
         // with the return values from reading and
         // writing. If either of these succeeds (ie,
         // returns 0), it will set err to 0 as well.
         
         // Read responses, if any.
         err = PrvFtpReadResponses (infoP);

         // Write command, if possible.
         // AND result with error code.
         err = PrvFtpWriteCommand (infoP) && err;
         
         // Read/write data as appropriate.
         // AND result with error code.
         err = PrvFtpDataExchange (infoP) && err;
      }

      // Check disconnections.
      // Update error - any success is a success.
      err = PrvFtpDisconnect (infoP) && err;
   }        
   
   // Move to next state, if appropriate.
   PrvFtpUpdate (infoP);

   // Return error.
   return err;
}

/*
   PnpFtpLogin:
   Login to the FTP server.
   Params:  infoP    FTP Session object.
            svrP     Name of server to login to.
            usrP     User name to login with, or NULL for anonymous.
            passP     Password for User account, or NULL for anonymous.
   Returns: 0        Success.
            < 0      Error.
   Note:    FTP Login Works as follows:

            1. Client Connects to Server on Port 21.
            2. Server sends Client a response.
            3. Client sends Server user name.
            4. Server sends Client a response.
            5. Client sends Server password.
            6. Server sends Client a response.

            Assuming that each of these responses indicated success
            (indicated by response from server starting with 2 or 3)
            we are now logged in.
            
    Note:   Changed for Chapter 12.
*/
Err
PnpFtpLogin (PnpFtpInfoPtr infoP, Char* svrP, Char* usrP, Char* passP)
{
   // Declare local variables.
   Err err = 0;
   Int16 resCode;

   // Make sure we are connected to the network.
   err = PrvFtpNetConnect (infoP);
   if (0 != err)
   {
      return err;
   }
                      
   // Retrieve internet address.
   err = PnpNetCalcAddress (&infoP->ctrlAddr, svrP, "ftp", "tcp");
   
   // Make sure we succeeded.
   if (0 != err)
   {
      return err;
   }
   
   // Enqueue Socket Connect.
   resCode = PrvFtpQueueCommand (infoP, kTranCodeSktConnect, NULL);
   
   // Make sure user name is valid.
   usrP = usrP ? usrP : (Char*) kDfltUser;
   
   // Queue USER command.
   resCode = PrvFtpQueueCommand (infoP, kTranCodeCmdUser, usrP);

   // Make sure password is valid.
   passP = passP ? passP : (Char*) kDfltPass;

   // Queue PASS command.
   resCode = PrvFtpQueueCommand (infoP, kTranCodeCmdPass, passP);

   // Return value.
   return err;
}

/*
   PnpFtpLogout:
   Logout from FTP Server.
   Params:  infoP    FTP Session object.
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpFtpLogout (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Int16 resCode;

   // Enqueue the QUIT command.
   resCode = PrvFtpQueueCommand (infoP, kTranCodeCmdQuit, NULL);

   // Enqueue the disconnect command.
   resCode = PrvFtpQueueCommand (infoP, kTranCodeSktDisconnect, NULL);
   
   // Return success.
   return 0;
}

/*
   PnpFtpCanLogin:
   Check if the we can login to an FTP server.
   Params:  infoP    FTP Session information.
   Returns: true     We are in a valid state for login.
            false    We are not in a valid state for login.
*/
Boolean
PnpFtpCanLogin (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Boolean canLogin = true;   // Return value.

   // Control socket must not be connected.
   if (kNilSocket != infoP->ctrlSocket)
   {
   // Control socket is connected.
   
      // Set return value to false.
      canLogin = false;
   }
   
   // Session must not be connecting.
   else if (kTranCodeSktConnect == infoP->tranCode)
   {
   // Session is connecting.
   // This will be the case from the time the session dequeues
   // the Connect transaction until PrvFtpConnect() calls
   // PnpNetConnect() to initiate the connection. The application
   // will handle one nilEvent during this time. This check
   // prevents the Login button from flashing at this time.
   
      // Set return value to false.
      canLogin = false;
   }

   // Return value.
   return canLogin;   
}

/*
   PnpFtpCanLogout:
   Check if the we can login to an FTP server.
   Params:  infoP    FTP Session information.
   Returns: true     We are in a valid state for login.
            false    We are not in a valid state for login.
*/
Boolean
PnpFtpCanLogout (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Boolean canLogout = true;   // Return value.
   
   // We must be connected.
   if (kNilSocket == infoP->ctrlSocket)
   {
   // We're not connected.
   
      // Set return value to false.
      canLogout = false;
   }
   
   // Transaction must be ready.
   else if (kTranCodeInactive != infoP->tranCode)
   {
   // Transaction is not ready.
   
      // Set return value to false.
      canLogout = false;
   }
                     
   // Return value.
   return canLogout;
}

/*
   PnpFtpList:
   List the contents of the server's current directory.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
*/
Err PnpFtpList (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Char portString [kInetAddrStringSize+1];
   
   // Connect.
   infoP->dataSocket = PnpNetSocketListen ();
      
   // Insist on success.
   ErrFatalDisplayIf  (kNilSocket == infoP->dataSocket,
                       "Unable to open listening socket!");      

   // Set state of data socket.
   infoP->dataState = kDataStateListening;

   // Calc. port address string.
   PrvFtpCalcPortAddress (infoP, portString);
      
   // Queue the PORT command.
   PrvFtpQueueCommand (infoP, kTranCodeCmdPort, portString);
      
   // Queue the LIST command.
   PrvFtpQueueCommand (infoP, kTranCodeCmdList, NULL);
   
   // Return value.
   return 0;
}

// --------------------------------------------------------------------
// Private function implementations
// --------------------------------------------------------------------

/*
   PrvFtpNetConnect:
   Connect to the network, if not already connected.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error.
*/
static Err
PrvFtpNetConnect (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;   // Return value.

   // Check if we are connected to the network.
   if (false == infoP->isNetConnect)
   {
   // We are not connected.
   
      // Connect to the network.
      err = PnpNetConnect ();
      if (!err)
      {
      // Successfully connected to network.

         // Set network connection flag to true.
         infoP->isNetConnect = true;
      }
   }
   
   // Return error.
   return err;
}

/*
   PrvFtpNetDisconnect:
   Disconnect from the network, if connected.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error.
*/
static Err
PrvFtpNetDisconnect (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;   // Return value.

   // Check if we are connected to the network.
   if (false != infoP->isNetConnect)
   {
   // We are connected.
   
      // Check if we can logout (this means we're logged in).
      // (We can't be logged in if we're not connected)
      if (kNilSocket != infoP->ctrlSocket)
      {
      // We're still logged in.

         // Logout.
         PnpFtpLogout (infoP);
      }

      // Disconnect from the network.
      err = PnpNetDisconnect ();

      // Clear network connection flag.
      infoP->isNetConnect = false;
   }
   
   // Return error.
   return err;
}

/*
   PrvFtpConnect:
   Move forward the Connect transaction, if appropriate.
   Params:  infoP    FTP Session information.
   Returns: 0        Transaction moved forward.
            <0       Transaction not moved forward or an error occurred.
            
    Note:   Changed for Chapter 12.
*/
Err
PrvFtpConnect(PnpFtpInfoPtr infoP)
{
   const Char* svrP = NULL;
   Char* textP = NULL;
   Err err = 0;

   // Make sure Connect transaction in Sending state.
   if (infoP->tranCode != kTranCodeSktConnect
       || infoP->tranState != kTranStateSending)
   {
   // We are not connecting or we are not in Sending state.
   
      // Return error.
      return -1;
   }
   
   // Check if the control socket is open.
   if (kNilSocket == infoP->ctrlSocket)
   {
      // Open the control socket.
      infoP->ctrlSocket = PnpNetSocketOpen (netSocketTypeStream);

      // Insist on success.
      ErrFatalDisplayIf (kNilSocket == infoP->ctrlSocket,
                         "Error opening socket.");
   }   

   // Attempt to connect with the server.
   err = PnpNetSocketConnect (infoP->ctrlSocket, &infoP->ctrlAddr);

   // Check result.
   if (0 == err)
   {
   // We are connected

      // Update transaction state.
      infoP->tranState = kTranStateReceiving;   // We are waiting for responses.

      // Clear transaction reply.
      infoP->tranReply = 0;
   
      // This is not really an error.
      err = 0;
   }
   else if (netErrWouldBlock != err)
   {
   // Error connecting socket.

      // Close the socket.
      PrvFtpCloseControl (infoP);
   }
   
   // Return value.
   return err;
}

Err
PrvFtpSelect (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   UInt16 width = 0;
   Err err = 0;
   
   // Initialize the file descriptor sets.
   netFDZero (&infoP->readFDs);
   netFDZero (&infoP->writeFDs);

   // Always check for reading.

   // Set Control Socket File Descriptor in Read Set.
   netFDSet (infoP->ctrlSocket, &infoP->readFDs);

   // Control socket bigger than width.
   width = infoP->ctrlSocket;

   // Check if we are writing a command.
   if (kTranStateSending == infoP->tranState 
       && kTranCodeCmdLast > infoP->tranCode)
   {
   // We are writing a command.

      // Set Control Socket File Descriptor in Write Set.
      netFDSet (infoP->ctrlSocket, &infoP->writeFDs);
   }

   // NEW!!
   // Check if we are receiving data.
   if (kDataStateReceiving == infoP->dataState)
   {
   // We are receiving data.
   
      // Set Data Socket File Descriptor in Read Set.
      netFDSet (infoP->dataSocket, &infoP->readFDs);
      
      // Update width if necessary.
      width = infoP->dataSocket >= width
              ? infoP->dataSocket + 1
              : width; 
   }
   
   // Call PnpNetSelect.
   err = PnpNetSelect (width+1, &infoP->readFDs, &infoP->writeFDs);
   
   // Return results.
   return err;
}

/*
   PrvFtpDisconnect:
   Disconnect from the FTP server. If we are not connected to a
   server, return an error.
   Params:  infoP    FTP Session Information.
   Returns: 0        Success.
            -1       Not connected to a server.
            -2       Error disconnecting from server.
*/
Err
PrvFtpDisconnect(PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1;
   
   // Check if we are doing a disconnect.
   if (infoP->tranCode != kTranCodeSktDisconnect)
   {
   // We are not disconnecting.
   
      // Return error.
      return -1;
   }
   
   // Check if we are in writing mode.
   if (infoP->tranState == kTranStateSending)
   {
   // We're in writing mode.
      
      // Add entry to log.
      PrvFtpAddLogItem (infoP, "DISCONNECT");
      
      // Attempt to shutdown connections to the server.
      PnpNetSocketShutdown (infoP->ctrlSocket);
      
      // Set transaction code and state.
      infoP->tranState = kTranStateReceiving;   // We are waiting for responses.
      
      // Set transaction reply.
      infoP->tranReply = 0;
      
      // Clear error.
      err = 0;
   }
   

   // Check if we are done reading responses.
   // If tran state is reading AND the last reply was -1, we did not
   //    receive any data during the last "read" transaction. We can
   //    assume all responses were read in response to QUIT command.
   // If tran state is success, we received a positive reply since
   //    we shut down. We can assume we are done receiving responses.
   // If tran state is failure, we received a negative reply since
   //    we shut down. We can assume we are done receiving responses.
   else if (kTranStateComplete == infoP->tranState
            || ( (kTranStateReceiving == infoP->tranState && -1 == infoP->tranReply)) )
   {
   // We are done reading responses.

      // Close the socket.
      PnpNetSocketDisconnect (infoP->ctrlSocket);
      infoP->ctrlSocket = kNilSocket;  // Control socket is no longer valid.

      // Update infoP's state.
      infoP->tranState = kTranStateComplete;

      // Set transaction reply to 0.
      infoP->tranReply = 0;
      
      // Clear error.
      err = 0;
   }

   // Return error.
   return err;
}

/*
   PrvFtpUpdate:
   Reviews transaction code and state and updates as appropriate.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
*/
static Err
PrvFtpUpdate (PnpFtpInfoPtr infoP)
{
   if (kTranStateComplete == infoP->tranState)
   {
   // Transaction complete.
   
      // What were we doing?
      switch (infoP->tranCode)
      {
         case kTranCodeCmdUser :
         case kTranCodeCmdPass :
         case kTranCodeCmdQuit :
         {
         // Login/out operation.

         // Check result.
            // See if transaction succeeded.
            if (!PrvFtpIsTranSuccess (infoP))
            {
            // Transaction failed and it wasn't a Disconnect.
            // For now, we disconnect on all failures or errors.

               // Clear all pending transactions.
               GQueClear (&infoP->tranQ);
               
               // Clear the input, output buffers.
               GBufEmpty (infoP->ctrlInBufH);
               GBufEmpty (infoP->ctrlOutBufH);

               // Enqueue the disconnect command.
               PrvFtpQueueCommand (infoP, kTranCodeSktDisconnect, NULL);
            }
            break;
         }
      }
      
      // Return session to inactive state.
      infoP->tranCode = kTranCodeInactive;
      infoP->tranState = kTranStateInactive;
   }
      
   // Check if we are ready to send new commands,
   //      and if there are new commands to send.
   if (kTranCodeInactive == infoP->tranCode
       && 0 < GQueNumEntries (&infoP->tranQ))
   {
   // We are ready to send and a command is pending.
   
      // Dequeue the command.
      infoP->tranCode = (PnpFtpTranCode) GQueDequeue (&infoP->tranQ);
      
      // Set the new transaction state.
      infoP->tranState = kTranStateSending;
  }
   
   // Return success.
   return 0;
}

/*
   PrvFtpIsTranSuccess:
   Check if the last transaction was a success.
   Params:  infoP    FTP Session information.
   Returns: true     Transaction succeeded.
            false    Transaction failed or an error occurred.
*/
static Boolean
PrvFtpIsTranSuccess (PnpFtpInfoPtr infoP)
{
   // Figure out range of positive replies.
   Int16 first = PrvFtpGetFirstPositiveReply (infoP),
         last = first + 99;
         
   // Transaction succeeded if reply code is in success range.
   return (first <= infoP->tranReply && last >= infoP->tranReply);
}

/*
   PrvFtpQueueCommand:
   Format and write a command to the FTP server.
   Params:  infoP    FTP Session information.
            tranCode   Transaction code of command to send.
            argP     Argument for command, NULL if none.
   Returns: 0        Success.
            <0       Error sending command.
*/
static Err
PrvFtpQueueCommand (PnpFtpInfoPtr infoP, PnpFtpTranCode tranCode, const Char* argP)
{
   // Declare local variables.
   Int16 size;
   Char* sendBufP = NULL;
   Char* cmdP = NULL;

   // Check if we are queuing a command transaction.   
   if (tranCode < kTranCodeCmdLast)
   {
   // It is a normal FTP transaction.
   
      // Get command from array.
      cmdP = pnpFtpTranNames [tranCode];
      
      // Calc. size of string.
      size = StrLen (cmdP) + sizeof (kCRLF); // sizeof includes nil-terminator.
      if (argP)
      {
      // There's an argument.
        
         // Add size of space and length of argument to size.
         size += sizeof (' ') + StrLen (argP);
         
         // Allocate the string.
         sendBufP = (Char*) MemPtrNew (size);
         
         // Format the string.
         size = StrPrintF (sendBufP, "%s %s%s", cmdP, argP, kCRLF);
      }
      else
      {
      // There's no argument.
        
         // Allocate the string.
         sendBufP = (Char*) MemPtrNew (size);
         
         // Format the string.
         size = StrPrintF (sendBufP, "%s%s", cmdP, kCRLF);
      }

      // Add command to the queue.
      GQueEnqueue (&infoP->tranQ, tranCode);
      
      // Append the command to the buffer.
      GBufAppend (infoP->ctrlOutBufH, (UInt8*) sendBufP, size);
      
      // Free the command.
      MemPtrFree (sendBufP), sendBufP = NULL;
   }

   // Check if we are queuing a socket connect transaction.   
   else
   {
   // We are queueing a disconnect.
   
      // Add disconnect command to the queue.
      GQueEnqueue (&infoP->tranQ, tranCode);
   }
         
   // Return success.
   return 0;
}

/*
   PrvFtpWriteCommand:
   Write the current command over the control connection to the FTP server.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error sending command or receiving response.
*/
static Err
PrvFtpWriteCommand (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Int16 leftToSend = 0;
   UInt16 totalSent = 0, sent = 0, sendLen = 0;
   const UInt8* sendP = NULL;
   Char* testP;
   Int16 testLen;
   Err err = 0;
   
   // Check if we are in write mode.
   if (kTranStateSending != infoP->tranState)
   {
   // We are not in write mode.
   
      // Return error.
      return pnpErrState;
   }
   
   // Check if we are in one of the "special" transaction codes.
   if (kTranCodeSktConnect == infoP->tranCode
       || kTranCodeSktDisconnect == infoP->tranCode)
   {
   // These are handled by PrvFtpConnect and PrvFtpDisconnect.
   
      // Return error.
      return -1;
   }
   
   // Check if the write FD is set for this socket.
   if (!netFDIsSet (infoP->ctrlSocket, &infoP->writeFDs))
   {
   // No data to read.
   
      // Return error.
      return -1;
   }

   // Get a read pointer into the buffer.
   sendP = GBufGetReadPtr (infoP->ctrlOutBufH, &sendLen);
   
   testP = pnpFtpTranNames [infoP->tranCode];
   testLen = StrLen (testP);
   
   // Make sure we are sending the right command.
   ErrFatalDisplayIf (0 != StrNCompare ((const Char*) sendP, testP, testLen),
                      "Queued command does not match command text.");
   
   // Look for CR-LF.
   leftToSend = PrvFtpFindCRLF (sendP, sendLen);
   
   // Insist that we find one!
   ErrFatalDisplayIf (leftToSend < 0,
                      "Improperly formatted FTP command!");
   
   // Add response to log.
   PrvFtpAddLogItem (infoP, (Char*) sendP);
   
   // Add 2 to length - this accounts for CR-LF.
   leftToSend += 2;
   
   while (leftToSend > 0 && !err)
   {
      // Assign to send size to send size.
      sendLen = leftToSend;
      
      // Write command.
      err = PnpNetSocketSend (infoP->ctrlSocket, (UInt8*) sendP, &sendLen);

      // Check for error.
      if (!err)
      {
         // Update buffer pointer.
         sendP += sendLen;
         
         // Update statistics.
         leftToSend -= sendLen;
         totalSent += sendLen;
      }
   }
   
   // Release the pointer.
   GBufReleaseReadPtr (infoP->ctrlOutBufH);
   
   // Check for success.
   if (totalSent > 0)
   {
   // Remove sent data from buffer.

      // Clear the buffer.
      GBufRead (infoP->ctrlOutBufH, NULL, totalSent);
      
      // Check if we have sent the entire command.
      if (0 == leftToSend) 
      {
      // We have sent the whole command.

         // Change state - now we wait for response.
         infoP->tranState = kTranStateReceiving;
      }
      
      // In any event, clear the error.
      err = 0;
   }
   
   // Return value.
   return err;
}

/*
   PrvFtpReadResponses:
   Read any available responses from the FTP server.
   Params:  infoP    FTP Session information.
   Returns: 0        Responses were read successfully.
            -1       No responses were read.
*/
static Err
PrvFtpReadResponses (PnpFtpInfoPtr infoP)
{
   const Int16 firstFailure = 400;
   const Int16 lastFailure  = 599;
   
   Err err = -1;
   Int16 firstSuccess, lastSuccess;
   Boolean complete = false;
   Int16 reply = 0;
   
   // Check if the read FD is set for this socket.
   if (!netFDIsSet (infoP->ctrlSocket, &infoP->readFDs))
   {
   // No data to read.
   
      // Return error.
      return -1;
   }

   // Check if we are in a reading state.
   if (kTranStateReceiving == infoP->tranState)
   {
   // We are waiting for a response.

      // Let's figure out range of replies to check.
      firstSuccess = PrvFtpGetFirstPositiveReply (infoP);
      lastSuccess = firstSuccess + 99;
   }
   
   else
   {
   // We don't care about the response.
   
      // Select widest range.
      firstSuccess = 0;
      lastSuccess = 399;
   }

   // Set last reply code to error.
   infoP->tranReply = -1;
   
   // Read any incoming data into buffers.
   PrvFtpReadControl (infoP);
 
   // Loop through, looking for replies.
   do
   {
      // Get next reply code.
      reply = PrvFtpReadResponse (infoP);
      
      // Check if reply code indicates success.
      if ( (reply >= firstSuccess && reply <= lastSuccess)
           || (reply >= firstFailure && reply <= lastFailure) )
      {
      // Reply indicates success for current transaction.
      
         // Save reply code for current transaction.
         infoP->tranReply = reply;
         
         // We have received a reply to the transaction.
         // We are done.
         complete = true;
         
         // Clear error.
         err = 0;
      }
      
   } while (0 <= reply);

   // Okay, we got our reply, see if we care.
   if (complete && kTranStateReceiving == infoP->tranState)
   {
   // We received a response and were waiting for it.
   
      // Set flag to success.
      infoP->tranState = kTranStateComplete;
   }

   // Return error.
   return err;
}

/*
   PrvFtpReadResponse:
   Read a response from the FTP server.
   Params:  infoP    FTP Session information.
   Returns: >0       Server response code.
            <0       Error receiving response.
*/
static Int16
PrvFtpReadResponse (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   Int16 retVal = 0, len = -1;
   UInt16 size = 0;
   const UInt8* bufP = NULL;
   Char* responseBufP = NULL;

   // Get read pointer to buffer.
   bufP = GBufGetReadPtr (infoP->ctrlInBufH, &size);

   // Check if there's anything in the buffer.
   if (NULL == bufP || 0 == size)
   {
   // No data to read.
   
      // Return error.
      return -1;
   }
   
   // Look for CRLF.
   len = PrvFtpFindCRLF (bufP, size);
   
   // Release read pointer.
   GBufReleaseReadPtr (infoP->ctrlInBufH);

   // Check if there is a CR-LF.
   if (0 > len)
   {
   // No CR-LF found.
   
      // Must not be entire response, wait for the rest.
      return -1;
   }
   
   // CR-LF found.
   
   // Get the response.
   size = len + sizeof kCRLF;   // Leave room for CR, LF and Nil-terminator.
   
   // Allocate a buffer.
   responseBufP = (Char*) MemPtrNew (size);
   
   // Make sure allocation succeeded.
   ErrFatalDisplayIf (NULL == responseBufP, "Memory allocation failed!");
   
   // Clear the buffer.
   MemSet (responseBufP, size, '\0');
   
   // Copy response to buffer.
   GBufRead (infoP->ctrlInBufH, (UInt8*) responseBufP, size-1);
   
   // Add response to log.
   PrvFtpAddLogItem (infoP, responseBufP);
   
   // Now check read numeric value of first three chars.
   responseBufP [3] = '\0';
   
   // Get response code.
   retVal = StrAToI (responseBufP);
   
   // Free the buffer.
   MemPtrFree (responseBufP);
   
   // Return value.
   return retVal;
}

/*
   PrvFtpReadControl:
   Read buffered responses from the control connection to the FTP server.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error sending command or receiving response.
*/
static Err
PrvFtpReadControl (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   UInt16 totalToRead = 0, totalRead = 0, read = 0;
   UInt8* bufP = NULL, * readP = NULL;
   
   // Figure out how big a buffer we need.
   totalToRead = GBufAvail (infoP->ctrlInBufH);
   
   // See if we can get anything.
   if (totalToRead == (UInt16) -1 || totalToRead == (UInt16) 0)
   {
   // Error or no room.
   
      // Return error.
      return -1;
   }
   
   // Allocate a buffer big enough for all of it.
   bufP = (UInt8*) MemPtrNew (totalToRead);
   
   // Clear the buffer.
   MemSet (bufP, totalToRead, '\0');
   
   // Initialize read pointer.
   readP = bufP;
   
   // Begin reading.
   do
   {
      // Set read amount to new value.
      read = totalToRead;
      
      // Receive initial response.
      err = PnpNetSocketReceive (infoP->ctrlSocket, readP, &read);

      // Check for error.
      if (!err)
      {
      // Success.
      
         // Check if anything was actually read.
         if (0 < read)
         {
         // Something was read.
         
            // Update buffer pointer.
            readP += read;
            
            // Update statistics.
            totalToRead -= read;
            totalRead += read;
         }
         else
         {
         // No data to read.
         
            // Set error.
            // (This will be cleared below if data has been received)
            err = -1;
         }
      }
   // Check if there is more to read or an error occurred.
   }  while (totalToRead > 0 && !err && false);
   
   // Check if anything read.
   if (0 < totalRead)
   {
   // Something was read.
   
      // Add this to buffer.
      GBufAppend (infoP->ctrlInBufH, bufP, totalRead);
   
      // Clear any errors.
      err = 0;   
   }
  
   // Don't forget to release the buffer.
   MemPtrFree (bufP);   bufP = NULL;
   
   // Return result.
   return err;
}

static Err
PrvFtpCloseControl (PnpFtpInfoPtr infoP)
{
   // Check if the control socket is open.
   if (kNilSocket != infoP->ctrlSocket)
   {
   // Socket is open.
   
      // Disconnect the socket.
      PnpNetSocketClose (infoP->ctrlSocket);
      infoP->ctrlSocket = kNilSocket;
   }
   
   // Clear all pending transactions.
   GQueClear (&infoP->tranQ);
   
   // Clear the input, output buffers.
   GBufEmpty (infoP->ctrlInBufH);
   GBufEmpty (infoP->ctrlOutBufH);

   // Set tranState to complete.
   infoP->tranState = kTranStateComplete;
   
   // Set tranReply to error.
   infoP->tranReply = -1;
   
   // Return success.
   return 0;
}

static Err
PrvFtpDataExchange (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1;
   
   // Check on data transfer state.
   switch (infoP->dataState)
   {
      case kDataStateInactive :
      {
      // Data is expected but nothing has been done.
      
         // Open socket in listening mode.
         err = PrvFtpDataListen (infoP);
         break;
      }
      
      case kDataStateListening :
      {
      // Listening for server connection.

         // Check for incoming connections.
         err = PrvFtpDataAccept (infoP);
         break;
      }
      
      case kDataStateReceiving :
      {
      // Receiving data from server.

         // Attempt to receive data.
         err = PrvFtpDataReceive (infoP);
         break;
      }
      
      case kDataStateComplete :
      {
      // Finished exchanging data with server.

         // Complete data transfer.
         err = PrvFtpDataComplete (infoP);
         break;
      }
   }
   
   // Return value.
   return err;
}

static Err
PrvFtpDataListen (PnpFtpInfoPtr infoP)
{
   // Connect.
   infoP->dataSocket = PnpNetSocketListen ();
      
   // Insist on success.
   ErrFatalDisplayIf  (kNilSocket == infoP->dataSocket,
                       "Unable to open listening socket!");      

   // Set state of data socket.
   infoP->dataState = kDataStateListening;
   
   // Return success.
   return 0;
}

static Err
PrvFtpDataAccept (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   NetSocketRef acceptSocket;
   Err err = 0;
   
   // Insist that data socket is valid.
   ErrFatalDisplayIf (kNilSocket == infoP->dataSocket,
                      "Data socket is not connected!");
                      
   // Attempt to accept connection.
   acceptSocket = PnpNetSocketAccept (infoP->dataSocket);
   
   // Check for success.
   if (kNilSocket != acceptSocket)
   {
   // Success.
   
      // Close listening socket.
      PnpNetSocketDisconnect (infoP->dataSocket);
      
      // Set data socket to new socket.
      infoP->dataSocket = acceptSocket;
      
      // Update data socket state.
      infoP->dataState = kDataStateReceiving;

      // Open database.
      GDtaOpen (&infoP->data, PNP_FILE_CREATOR, pnpListDBName);
   }
   
   else
   {
   // Error.
   
      // Set error code.
      err = -1;
   }
   
   // Return error.
   return err;
}

static Err
PrvFtpDataReceive (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1, retVal = -1;
   
   GDtaAppendRecord (&infoP->data, 1024);
   
   // Insist that data socket is valid.
   ErrFatalDisplayIf (kNilSocket == infoP->dataSocket,
                      "Data socket is not connected!");

   do 
   {
      const UInt16 len = infoP->data.recordLength;
      err = PnpNetSocketDmReceive (infoP->dataSocket, &infoP->data);
      if (!err)
      {                                          
         retVal = 0;
         if (len == infoP->data.recordLength)
         {
         // Record length has not changed.
         // This means remote is done sending.
         // Otherwise there would have been an error.
         
         // We're done receiving data.
         
            // Set new state.
            infoP->dataState = kDataStateComplete;
            err = -1;
         }
      }
   }  while (!err);
   
   // Release record.
   GDtaReleaseRecord (&infoP->data);
   
   // Return error.
   return retVal;
}
   
static Err
PrvFtpDataComplete (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   
   // Insist that data socket is valid.
   ErrFatalDisplayIf (kNilSocket == infoP->dataSocket,
                      "Data socket is not connected!");

   // Close the database.
   GDtaClose (&infoP->data);
   
   // Shut down data connection.
   PnpNetSocketShutdown (infoP->dataSocket);
   
   // Close data connection.
   PnpNetSocketDisconnect (infoP->dataSocket);
      
   // Data connection is invalid.
   infoP->dataSocket = kNilSocket;
      
   // Update log.
   PrvFtpAddLogItem (infoP, "DISCONNECT DATA");

   // Update data state.
   infoP->dataState = kDataStateNoTransfer;
      
   // Return error.
   return 0;
}

static Int16
PrvFtpGetFirstPositiveReply (PnpFtpInfoPtr infoP)
{
   return (infoP->tranCode != kTranCodeCmdUser) ? 200 : 300;
}

static Err
PrvFtpCalcPortAddress (PnpFtpInfoPtr infoP, Char* portString)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   UInt16 port, hiPort, loPort;
   Char* chP;
   
   // Get dotted decimal IP address.
   err = PnpNetSocketAddress (infoP->dataSocket, portString);
   
   // Insist on success.
   ErrFatalDisplayIf (err, "Unable to retrieve socket address!");
   
   // Convert inet address.
   for (chP = portString; (*chP) != '\0'; chP++)
   {
      // Check if character is a digit.
      if (('0' > (*chP)) || ('9' < (*chP)))
      {
      // Character is not a digit.
      
         // Replace w/ a comma.
         (*chP) = ',';
      }
   }
   
   // Append port.
   err = PnpNetSocketPort (infoP->dataSocket, &port);
   
   // Insist on success.
   ErrFatalDisplayIf (err, "Unable to retrieve socket port number!");
   
   chP = portString + StrLen (portString);
   loPort = (NetNToHS (port) & 0x00FF);
   hiPort = (NetNToHS (port) & 0xFF00) >> 8;
   StrPrintF (chP, ",%d,%d", hiPort, loPort);
   
   // Return value.
   return err;
}

static Err
PrvFtpAddLogItem (PnpFtpInfoPtr infoP, Char* logItem)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   Char* bufP;
   Int16 len, itemLen;
   Char* testP = pnpFtpTranNames [kTranCodeCmdQuit];

   // Debug - let's look for cmd quit.
   if (0 == StrNCompare (logItem, testP, StrLen (testP))) 
   {
   // Sending QUIT command.
      err = 0;
   }
   
   // First let's get the log item string length.
   itemLen = StrLen (logItem);
   
   // Now check if there's a CRLF before the end.
   len = PrvFtpFindCRLF ((UInt8*) logItem, itemLen);
   if (0 > len)
   {
   // There was no CRLF.
   
      // Set length to length of string.
      len = itemLen;
   }
   
   // Allocate a temp block of memory.
   bufP = MemPtrNew (len+1);

   // Check if allocation succeeded.
   if (!bufP)
   {
   // Memory allocation failed.

      // Return approp. error.
      return -1;
   }

   // Copy log message to buffer.
   StrNCopy (bufP, logItem, len);
   bufP [len] = '\0';

   // Add result to log.
   err = PnpLogAddItem (pnpLogFtp, bufP);

   // Free the memory.
   MemPtrFree (bufP);

   // Return error.
   return err;
}
   
static Int16
PrvFtpFindCRLF (const UInt8* bufP, UInt16 size)
{
   // Declare local variables.
   Int16 i, len = -1;
   
   // Traverse read pointer looking for first CRLF.
   for (i = 0; -1 == len && i < size-1; i++)
   {
      if (bufP [i] == 0x0D && bufP [i+1] == 0x0A)
      {
         // Set length to i.
         len = i;
      }
   }
   
   // Return value.
   return len;
}

